@@ -131,7 +131,7 @@ wl_array_add(struct wl_array *array, size_t size)
 		array->alloc = alloc;
 	}
 
-	p = array->data + array->size;
+	p = (char *)array->data + array->size;
 	array->size += size;
 
 	return p;
@@ -147,7 +147,9 @@ wl_array_copy(struct wl_array *array, struct wl_array *source)
 		array->size = source->size;
 	}
 
-	memcpy(array->data, source->data, source->size);
+	if (source->size > 0)
+		memcpy(array->data, source->data, source->size);
+
 	return 0;
 }
 
@@ -195,7 +197,6 @@ wl_map_insert_new(struct wl_map *map, uint32_t flags, void *data)
 	union map_entry *start, *entry;
 	struct wl_array *entries;
 	uint32_t base;
-	uint32_t count;
 
 	if (map->side == WL_MAP_CLIENT_SIDE) {
 		entries = &map->client_entries;
@@ -216,25 +217,10 @@ wl_map_insert_new(struct wl_map *map, uint32_t flags, void *data)
 		start = entries->data;
 	}
 
-	/* wl_array only grows, so if we have too many objects at
-	 * this point there's no way to clean up. We could be more
-	 * pro-active about trying to avoid this allocation, but
-	 * it doesn't really matter because at this point there is
-	 * nothing to be done but disconnect the client and delete
-	 * the whole array either way.
-	 */
-	count = entry - start;
-	if (count > WL_MAP_MAX_OBJECTS) {
-		/* entry->data is freshly malloced garbage, so we'd
-		 * better make it a NULL so wl_map_for_each doesn't
-		 * dereference it later. */
-		entry->data = NULL;
-		return 0;
-	}
 	entry->data = data;
 	entry->next |= (flags & 0x1) << 1;
 
-	return count + base;
+	return (entry - start) + base;
 }
 
 int
@@ -251,9 +237,6 @@ wl_map_insert_at(struct wl_map *map, uint32_t flags, uint32_t i, void *data)
 		i -= WL_SERVER_ID_START;
 	}
 
-	if (i > WL_MAP_MAX_OBJECTS)
-		return -1;
-
 	count = entries->size / sizeof *start;
 	if (count < i)
 		return -1;
@@ -288,9 +271,6 @@ wl_map_reserve_new(struct wl_map *map, uint32_t i)
 		i -= WL_SERVER_ID_START;
 	}
 
-	if (i > WL_MAP_MAX_OBJECTS)
-		return -1;
-
 	count = entries->size / sizeof *start;
 
 	if (count < i)
@@ -383,18 +363,21 @@ wl_map_lookup_flags(struct wl_map *map, uint32_t i)
 static enum wl_iterator_result
 for_each_helper(struct wl_array *entries, wl_iterator_func_t func, void *data)
 {
-	union map_entry *start, *end, *p;
 	enum wl_iterator_result ret = WL_ITERATOR_CONTINUE;
+	union map_entry entry, *start;
+	size_t count;
 
-	start = entries->data;
-	end = (union map_entry *) ((char *) entries->data + entries->size);
+	start = (union map_entry *) entries->data;
+	count = entries->size / sizeof(union map_entry);
 
-	for (p = start; p < end; p++)
-		if (p->data && !map_entry_is_free(*p)) {
-			ret = func(map_entry_get_data(*p), data, map_entry_get_flags(*p));
+	for (size_t idx = 0; idx < count; idx++) {
+		entry = start[idx];
+		if (entry.data && !map_entry_is_free(entry)) {
+			ret = func(map_entry_get_data(entry), data, map_entry_get_flags(entry));
 			if (ret != WL_ITERATOR_CONTINUE)
 				break;
 		}
+	}
 
 	return ret;
 }
